"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const GethBin_1 = require("./GethBin");
const Web3_1 = require("./Web3");
const net_1 = require("net");
const event = require("./Constants");
const events_1 = require("events");
const child_process_1 = require("child_process");
const Promise = require("bluebird");
const os_1 = require("os");
const path_1 = require("path");
const platform = os_1.type();
const symbolEnforcer = Symbol('go-ethereum');
const symbol = Symbol('go-ethereum');
const cpuPriority = {
    unix: {
        [event.PriorityCode.HIGH]: '-17',
        [event.PriorityCode.NORMAL]: '0',
        [event.PriorityCode.LOW]: '17',
    },
    win: {
        [event.PriorityCode.HIGH]: 'high',
        [event.PriorityCode.NORMAL]: 'normal',
        [event.PriorityCode.LOW]: 'below normal',
    },
};
class GethConnector extends events_1.EventEmitter {
    constructor(enforcer) {
        super();
        this.downloadManager = new GethBin_1.GethBin();
        this.ipcStream = new Web3_1.Web3();
        this.logger = console;
        this.spawnOptions = new Map();
        this.serviceStatus = {
            process: false,
            api: false,
            version: '',
        };
        this.watchers = new Map();
        this.connectedToLocal = false;
        this.cpuPriority = event.PriorityCode.LOW;
        this._downloadEventsEnabled = false;
        if (enforcer !== symbolEnforcer) {
            throw new Error('Use .getInstance() instead of new constructor');
        }
    }
    get web3() {
        return this.ipcStream.web3;
    }
    static getInstance() {
        if (!this[symbol]) {
            this[symbol] = new GethConnector(symbolEnforcer);
        }
        return this[symbol];
    }
    static getDefaultDatadir() {
        let dataDir;
        switch (platform) {
            case 'Linux':
                dataDir = path_1.join(os_1.homedir(), '.ethereum');
                break;
            case 'Darwin':
                dataDir = path_1.join(os_1.homedir(), 'Library', 'Ethereum');
                break;
            case 'Windows_NT':
                dataDir = path_1.join(process.env.APPDATA, '/Ethereum');
                break;
            default:
                throw new Error('Platform not supported');
        }
        return dataDir;
    }
    static getDefaultIpcPath() {
        const dataDirPath = GethConnector.getDefaultDatadir();
        let ipcPath;
        switch (platform) {
            case 'Windows_NT':
                ipcPath = '\\\\.\\pipe\\geth.ipc';
                break;
            default:
                ipcPath = path_1.join(dataDirPath, 'geth.ipc');
                break;
        }
        return ipcPath;
    }
    setLogger(logger) {
        this.logger = logger;
    }
    setBinPath(path) {
        this.downloadManager = new GethBin_1.GethBin(path);
        this._downloadEventsEnabled = false;
    }
    start(options) {
        this.emit(event.STARTING);
        this.setOptions(options);
        this.socket = new net_1.Socket();
        return this._checkBin().then((binPath) => {
            if (!binPath) {
                this.emit(event.FAILED, 'gethBin');
                this.serviceStatus.process = false;
                return false;
            }
            if (this.connectedToLocal) {
                return false;
            }
            this.gethService = child_process_1.spawn(binPath, this._flattenOptions(), { detached: true });
            return true;
        }).then((passed) => {
            if (passed) {
                this._attachEvents();
            }
            return passed;
        });
    }
    stop() {
        this.emit(event.STOPPING);
        this._flushEvents();
        if (!this.gethService) {
            this.emit(event.STOPPED);
        }
        const killProcess = (this.gethService) ? this.gethService.kill() : true;
        return Promise.delay(500);
    }
    getChainFolder() {
        return this.web3.admin.getDatadir().then((datadir) => {
            return path_1.join(datadir, 'chaindata');
        });
    }
    connectToLocal() {
        this._connectToIPC();
        this.connectedToLocal = true;
    }
    setCpuPriority(level, immediate) {
        if (!event.PriorityCode.hasOwnProperty(level)) {
            throw new Error('Invalid priority level');
        }
        this.cpuPriority = level;
        if (immediate) {
            this.executeCpuPriority();
        }
    }
    executeCpuPriority() {
        if (!this.gethService.pid) {
            this.logger.warn('cpu:geth process is not started');
            return;
        }
        const command = (process.platform === 'win32') ?
            `wmic process where name="geth.exe" CALL setpriority "${cpuPriority.win[this.cpuPriority]}"` :
            `renice -n ${cpuPriority.unix[this.cpuPriority]} -p ${this.gethService.pid}`;
        child_process_1.exec(command, (error, stdout, stderr) => {
            if (error) {
                this.logger.error(`cpu:geth:exec error: ${error}`);
                return;
            }
            this.logger.debug(`cpu:geth:stdout: ${stdout}`);
            this.logger.debug(`cpu:geth:stderr: ${stderr}`);
        });
    }
    getCpuPriority() {
        return this.cpuPriority;
    }
    enableDownloadEvents() {
        if (this._downloadEventsEnabled) {
            return;
        }
        this.downloadManager.wrapper.downloadProgress.on(event.DOWNLOAD_STARTED, () => {
            this.emit(event.DOWNLOAD_STARTED);
        });
        this.downloadManager.wrapper.downloadProgress.on(event.DOWNLOAD_PROGRESS, (progress) => {
            this.emit(event.DOWNLOAD_PROGRESS, progress);
        });
        this.downloadManager.wrapper.downloadProgress.on(event.DOWNLOAD_ERROR, (error) => {
            this.emit(event.DOWNLOAD_ERROR, error);
        });
        this._downloadEventsEnabled = true;
    }
    setOptions(options) {
        let localOptions;
        if (options) {
            if (options.hasOwnProperty('cpu')) {
                this.setCpuPriority(options.cpu);
                delete options.cpu;
            }
            if (platform === 'Windows_NT' && options.hasOwnProperty('ipcpath')) {
                options.ipcpath = path_1.join('\\\\.\\pipe', options.ipcpath);
            }
            if (platform !== 'Windows_NT' && options.hasOwnProperty('datadir') && !options.hasOwnProperty('ipcpath')) {
                options.ipcpath = path_1.join(options.datadir, 'geth.ipc');
            }
        }
        if (this.spawnOptions.size) {
            if (!options) {
                return this.spawnOptions;
            }
            localOptions = options;
        }
        else {
            localOptions = Object.assign({
                datadir: GethConnector.getDefaultDatadir(),
                ipcpath: GethConnector.getDefaultIpcPath(),
            }, event.START_OPTIONS, options);
        }
        for (let option in localOptions) {
            if (localOptions.hasOwnProperty(option)) {
                this.spawnOptions.set(option, localOptions[option]);
            }
        }
        if (this.spawnOptions.get(event.BIN_PATH)) {
            this.setBinPath(this.spawnOptions.get(event.BIN_PATH));
            this.spawnOptions.delete(event.BIN_PATH);
        }
        return this.spawnOptions;
    }
    restart(waitTime = 5000) {
        return Promise
            .resolve(this.stop())
            .delay(waitTime)
            .then(() => this.start());
    }
    writeGenesis(genesisPath, cb) {
        this._checkBin().then((binPath) => {
            if (binPath) {
                const dataDir = (this.spawnOptions.get('datadir')) ? this.spawnOptions.get('datadir') : '';
                let command = (dataDir) ? `--datadir "${dataDir}"` : '';
                command += ` init "${genesisPath}"`;
                child_process_1.exec(`"${binPath}" ${command}`, (error, stdout) => {
                    cb(error, stdout);
                });
            }
        });
    }
    _flushEvents() {
        this.web3.reset();
        this.serviceStatus.api = false;
        this.serviceStatus.process = false;
        this.serviceStatus.version = '';
        if (this.socket) {
            this.socket.removeAllListeners();
        }
        if (!this.connectedToLocal) {
            if (this.watchers.get(event.START_FILTER)) {
                this.gethService
                    .stderr
                    .removeListener('data', this.watchers.get(event.START_FILTER));
                this.watchers
                    .delete(event.START_FILTER);
            }
            if (this.watchers.get(event.INFO_FILTER)) {
                this.gethService
                    .stdout
                    .removeListener('data', this.watchers.get(event.INFO_FILTER));
                this.gethService
                    .stderr
                    .removeListener('data', this.watchers.get(event.INFO_FILTER));
                this.watchers
                    .delete(event.INFO_FILTER);
            }
        }
        if (this.socket) {
            this.socket.end();
        }
        this.socket = null;
        return true;
    }
    _checkBin() {
        return new Promise((resolve, reject) => {
            this.downloadManager.check((err, data) => {
                if (err) {
                    this.logger.error(err);
                    this.emit(event.BINARY_CORRUPTED, err);
                    this.downloadManager.deleteBin();
                    return reject(err);
                }
                if (data.binPath) {
                    return resolve(data.binPath);
                }
            });
        });
    }
    _flattenOptions() {
        const arrayOptions = [];
        for (let [key, value] of this.spawnOptions) {
            arrayOptions.push(`--${key}`);
            if (value) {
                arrayOptions.push(value);
            }
        }
        return arrayOptions;
    }
    _attachEvents() {
        this.__listenProcess();
        this._watchGethStd();
        this._attachListeners();
    }
    __listenProcess() {
        this.gethService.once('exit', (code, signal) => {
            let message;
            if (code) {
                message = `geth: exited with code: ${code}`;
                this.logger.error(message);
                this.emit(event.ERROR, message);
            }
            else {
                message = `geth: received signal: ${signal}`;
                this.logger.debug(message);
            }
            this._flushEvents();
            this.serviceStatus.process = false;
        });
        this.gethService.once('close', (code, signal) => {
            this.logger.debug('geth:spawn:close:', code, signal);
            this.emit(event.STOPPED);
        });
        this.gethService.once('error', (code) => {
            this.logger.error(`geth:spawn:error: ${code}`);
            this.serviceStatus.process = false;
            this.emit(event.FAILED, 'gethService');
        });
    }
    _tailGethLog() {
        if (this.watchers.get(event.START_FILTER)) {
            this.gethService.stderr.removeListener('data', this.watchers.get(event.START_FILTER));
            this.watchers.delete(event.START_FILTER);
        }
        const infoFilter = (data) => {
            this.logger.debug(data.toString());
        };
        this.watchers.set(event.INFO_FILTER, infoFilter);
        this.gethService.stdout.on('data', this.watchers.get(event.INFO_FILTER));
        this.gethService.stderr.on('data', this.watchers.get(event.INFO_FILTER));
    }
    _watchGethStd() {
        let started = false;
        this.serviceStatus.process = false;
        const timeout = setTimeout(() => {
            this.gethService.stderr.removeAllListeners('data');
            this.emit(event.ERROR, 'geth connection timeout');
        }, 20000);
        const startFilter = (data) => {
            let log = data.toString();
            if (log.includes('clock seems off')) {
                this.emit(event.TIME_NOT_SYNCED, log);
            }
            if (log.includes('Fatal') ||
                log.includes('Synchronisation failed')) {
                this.logger.error(`geth:sync:error: ${log}`);
                this.emit(event.FATAL, log);
                clearTimeout(timeout);
            }
            if (log.includes('IPC endpoint opened')) {
                this.serviceStatus.process = true;
                clearTimeout(timeout);
                this._connectToIPC();
            }
            this.logger.debug(log);
        };
        this.watchers.set(event.START_FILTER, startFilter);
        this.gethService.stderr.on('data', this.watchers.get(event.START_FILTER));
    }
    _attachListeners() {
        this.removeListener(event.STARTED, this._tailGethLog);
        this.removeListener(event.FATAL, this.socket.end);
        this.once(event.STARTED, this._tailGethLog);
        this.once(event.FATAL, () => this.socket ? this.socket.end : null);
    }
    _connectToIPC() {
        this.ipcStream.setProvider(this.spawnOptions.get('ipcpath'), this.socket);
        this.socket.once('connect', () => {
            this.logger.info('Connection to go-ethereum ipc established!');
            this.emit(event.IPC_CONNECTED);
            this._checkRunningSevice().then((status) => {
                if (status) {
                    this.serviceStatus.api = true;
                }
            });
            this.emit(event.STARTED);
        });
        this.socket.once('error', (error) => {
            this.web3.reset();
            this.logger.error(error.message);
            this.serviceStatus.api = false;
            this.emit(event.IPC_DISCONNECTED);
        });
    }
    _checkRunningSevice() {
        const requiredVersion = GethBin_1.GethBin.requiredVersion();
        const runningServiceProps = [
            this.web3.version.getNodeAsync(),
            this.web3.version.getNetworkAsync(),
        ];
        return Promise.all(runningServiceProps).then(([buildVersion, networkId]) => {
            let message;
            networkId = parseInt(networkId, 10);
            this.serviceStatus.version = buildVersion;
            if (!buildVersion.includes(requiredVersion)) {
                message = `required geth version: ${requiredVersion}, found: ${buildVersion}, updating executable...`;
                this.logger.warn(message);
                this.emit(event.UPDATING_BINARY, message);
                return this.stop()
                    .delay(5000)
                    .then(() => this.downloadManager.deleteBin())
                    .then(() => this.start());
            }
            let netId = this.spawnOptions.get('networkid');
            if (this.spawnOptions.has('testnet')) {
                netId = 2;
            }
            if (netId && networkId !== netId) {
                message = `required ethereum network: ${netId}, found: ${networkId}`;
                this.logger.error(message);
                this.emit(event.FATAL, message);
                return false;
            }
            this.emit(event.ETH_NODE_OK);
            return true;
        });
    }
}
exports.default = GethConnector;
